@@ -58,13 +58,13 @@ static int oc_vasprintf(char **strp, const char *fmt, va_list ap)
 	   strings in openconnect. */
 #ifdef HAVE_VA_COPY
 	va_copy(ap2, ap);
-#elif defined (HAVE___VA_COPY)
+#elif defined(HAVE___VA_COPY)
 	__va_copy(ap2, ap);
 #else
 #error No va_copy()!
-	// You could try this.
+	/* You could try this. */
 	ap2 = ap;
-	// Or this
+	/* Or this */
 	*ap2 = *ap;
 #endif
 	len = vsnprintf(res, 160, fmt, ap2);
@@ -77,7 +77,7 @@ static int oc_vasprintf(char **strp, const char *fmt, va_list ap)
 		res = NULL;
 		goto err;
 	}
-	if (len >=0 && len < 160)
+	if (len >= 0 && len < 160)
 		goto out;
 
 	free(res);
@@ -131,7 +131,7 @@ ssize_t openconnect__getline(char **lineptr, size_t *n, FILE *stream)
 			break;
 
 		*n *= 2;
-		*lineptr = realloc(*lineptr, *n);
+		realloc_inplace(*lineptr, *n);
 		if (!*lineptr)
 			return -1;
 	}
@@ -140,3 +140,25 @@ ssize_t openconnect__getline(char **lineptr, size_t *n, FILE *stream)
 	return -1;
 }
 #endif
+
+#ifndef HAVE_STRCASESTR
+#include <ctype.h>
+
+char *openconnect__strcasestr(const char *haystack, const char *needle)
+{
+	int hlen = strlen(haystack);
+	int nlen = strlen(needle);
+	int i, j;
+
+	for (i = 0; i < hlen - nlen + 1; i++) {
+		for (j = 0; j < nlen; j++) {
+			if (tolower(haystack[i + j]) !=
+			    tolower(needle[j]))
+				break;
+		}
+		if (j == nlen)
+			return (char *)haystack + i;
+	}
+	return NULL;
+}
+#endif
