@@ -119,7 +119,8 @@ static void calculate_mtu(struct openconnect_info *vpninfo, int *base_mtu, int *
 			vpn_progress(vpninfo, PRG_TRACE,
 				     _("TCP_INFO rcv mss %d, snd mss %d, adv mss %d, pmtu %d\n"),
 				     ti.tcpi_rcv_mss, ti.tcpi_snd_mss, ti.tcpi_advmss, ti.tcpi_pmtu);
-			if (!*base_mtu) *base_mtu = ti.tcpi_pmtu;
+			if (!*base_mtu)
+				*base_mtu = ti.tcpi_pmtu;
 			if (!*mtu) {
 				if (ti.tcpi_rcv_mss < ti.tcpi_snd_mss)
 					*mtu = ti.tcpi_rcv_mss - 13;
@@ -144,6 +145,8 @@ static void calculate_mtu(struct openconnect_info *vpninfo, int *base_mtu, int *
 		/* Default */
 		*mtu = 1406;
 	}
+	if (*mtu < 1280)
+		*mtu = 1280;
 }
 
 static int start_cstp_connection(struct openconnect_info *vpninfo)
@@ -169,7 +172,7 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 	vpninfo->vpn_domain = vpninfo->vpn_proxy_pac = NULL;
 	vpninfo->banner = NULL;
 
-	for (i=0; i<3; i++)
+	for (i = 0; i < 3; i++)
 		vpninfo->vpn_dns[i] = vpninfo->vpn_nbns[i] = NULL;
 
 	for (inc = vpninfo->split_includes; inc; ) {
@@ -213,12 +216,12 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 		buf_append(buf, sizeof(buf), "X-CSTP-Base-MTU: %d\r\n", base_mtu);
 	buf_append(buf, sizeof(buf), "X-CSTP-MTU: %d\r\n", mtu);
 	buf_append(buf, sizeof(buf), "X-CSTP-Address-Type: %s\r\n",
-			       vpninfo->disable_ipv6?"IPv4":"IPv6,IPv4");
+			       vpninfo->disable_ipv6 ? "IPv4" : "IPv6,IPv4");
 	buf_append(buf, sizeof(buf), "X-DTLS-Master-Secret: ");
 	for (i = 0; i < sizeof(vpninfo->dtls_secret); i++)
 		buf_append(buf, sizeof(buf), "%02X", vpninfo->dtls_secret[i]);
 	buf_append(buf, sizeof(buf), "\r\nX-DTLS-CipherSuite: %s\r\n\r\n",
-			       vpninfo->dtls_ciphers?:"AES256-SHA:AES128-SHA:DES-CBC3-SHA:DES-CBC-SHA");
+			       vpninfo->dtls_ciphers ? : "AES256-SHA:AES128-SHA:DES-CBC3-SHA:DES-CBC-SHA");
 
 	openconnect_SSL_write(vpninfo, buf, strlen(buf));
 
@@ -302,10 +305,17 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 
 		if (!new_option->option || !new_option->value) {
 			vpn_progress(vpninfo, PRG_ERR, _("No memory for options\n"));
+			free(new_option->option);
+			free(new_option->value);
+			free(new_option);
 			return -ENOMEM;
 		}
 
-		vpn_progress(vpninfo, PRG_TRACE, "%s: %s\n", buf, colon);
+		/* This contains the whole document, including the webvpn cookie. */
+		if (!strcasecmp(buf, "X-CSTP-Post-Auth-XML"))
+			vpn_progress(vpninfo, PRG_TRACE, "%s: %s\n", buf, _("<elided>"));
+		else
+			vpn_progress(vpninfo, PRG_TRACE, "%s: %s\n", buf, colon);
 
 		if (!strncmp(buf, "X-DTLS-", 7)) {
 			*next_dtls_option = new_option;
@@ -600,9 +610,8 @@ static int inflate_and_queue_packet(struct openconnect_info *vpninfo,
 	pkt_sum = buf[len - 1] | (buf[len - 2] << 8) |
 		(buf[len - 3] << 16) | (buf[len - 4] << 24);
 
-	if (vpninfo->inflate_adler32 != pkt_sum) {
+	if (vpninfo->inflate_adler32 != pkt_sum)
 		vpninfo->quit_reason = "Compression (inflate) adler32 failure";
-	}
 
 	vpn_progress(vpninfo, PRG_TRACE,
 		     _("Received compressed data packet of %ld bytes\n"),
@@ -612,7 +621,7 @@ static int inflate_and_queue_packet(struct openconnect_info *vpninfo,
 	return 0;
 }
 
-#if defined (OPENCONNECT_OPENSSL)
+#if defined(OPENCONNECT_OPENSSL)
 static int cstp_read(struct openconnect_info *vpninfo, void *buf, int maxlen)
 {
 	int len, ret;
@@ -654,7 +663,7 @@ static int cstp_write(struct openconnect_info *vpninfo, void *buf, int buflen)
 		return -1;
 	}
 }
-#elif defined (OPENCONNECT_GNUTLS)
+#elif defined(OPENCONNECT_GNUTLS)
 static int cstp_read(struct openconnect_info *vpninfo, void *buf, int maxlen)
 {
 	int ret;
@@ -706,7 +715,7 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	   we should probably remove POLLIN from the events we're looking for,
 	   and add POLLOUT. As it is, though, it'll just chew CPU time in that
 	   fairly unlikely situation, until the write backlog clears. */
-	while ( (len = cstp_read(vpninfo, buf, sizeof(buf))) > 0) {
+	while ((len = cstp_read(vpninfo, buf, sizeof(buf))) > 0) {
 		int payload_len;
 
 		if (buf[0] != 'S' || buf[1] != 'T' ||
@@ -725,7 +734,7 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 			continue;
 		}
 		vpninfo->ssl_times.last_rx = time(NULL);
-		switch(buf[6]) {
+		switch (buf[6]) {
 		case AC_PKT_DPD_OUT:
 			vpn_progress(vpninfo, PRG_TRACE,
 				     _("Got CSTP DPD request\n"));
@@ -959,10 +968,10 @@ int cstp_bye(struct openconnect_info *vpninfo, const char *reason)
 	int reason_len;
 
 	/* already lost connection? */
-#if defined (OPENCONNECT_OPENSSL)
+#if defined(OPENCONNECT_OPENSSL)
 	if (!vpninfo->https_ssl)
 		return 0;
-#elif defined (OPENCONNECT_GNUTLS)
+#elif defined(OPENCONNECT_GNUTLS)
 	if (!vpninfo->https_sess)
 		return 0;
 #endif
