@@ -59,7 +59,8 @@ unsigned char unhex(const char *data)
 time_t time(time_t *t)
 {
 	time_t x = 0x3ab2d948;
-	if (t) *t = x;
+	if (t)
+		*t = x;
 	return x;
 }
 
@@ -105,14 +106,14 @@ int RAND_bytes(char *buf, int len)
  * their clients use anyway.
  */
 
-#if defined (DTLS_OPENSSL)
+#if defined(DTLS_OPENSSL)
 #define DTLS_SEND SSL_write
 #define DTLS_RECV SSL_read
 
 #ifdef HAVE_DTLS1_STOP_TIMER
 /* OpenSSL doesn't deliberately export this, but we need it to
    workaround a DTLS bug in versions < 1.0.0e */
-extern void dtls1_stop_timer (SSL *);
+extern void dtls1_stop_timer(SSL *);
 #endif
 
 static int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
@@ -252,32 +253,32 @@ int dtls_try_handshake(struct openconnect_info *vpninfo)
 #if OPENSSL_VERSION_NUMBER >= 0x1000005fL
 		/* OpenSSL 1.0.0e or above doesn't resend anyway; do nothing.
 		   However, if we were *built* against 1.0.0e or newer, but at
-		   runtime we find that we are being run against an older 
+		   runtime we find that we are being run against an older
 		   version, warn about it. */
 		if (SSLeay() < 0x1000005fL) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Your OpenSSL is older than the one you built against, so DTLS may fail!"));
 		}
-#elif defined (HAVE_DTLS1_STOP_TIMER)
+#elif defined(HAVE_DTLS1_STOP_TIMER)
 		/*
 		 * This works for any normal OpenSSL that supports
 		 * Cisco DTLS compatibility (0.9.8m to 1.0.0d inclusive,
 		 * and even later versions although it isn't needed there.
 		 */
 		dtls1_stop_timer(vpninfo->dtls_ssl);
-#elif defined (BIO_CTRL_DGRAM_SET_NEXT_TIMEOUT)
+#elif defined(BIO_CTRL_DGRAM_SET_NEXT_TIMEOUT)
 		/*
 		 * Debian restricts visibility of dtls1_stop_timer()
 		 * so do it manually. This version also works on all
 		 * sane versions of OpenSSL:
 		 */
-		memset (&(vpninfo->dtls_ssl->d1->next_timeout), 0,
-			sizeof((vpninfo->dtls_ssl->d1->next_timeout)));
+		memset(&(vpninfo->dtls_ssl->d1->next_timeout), 0,
+		       sizeof((vpninfo->dtls_ssl->d1->next_timeout)));
 		vpninfo->dtls_ssl->d1->timeout_duration = 1;
 		BIO_ctrl(SSL_get_rbio(vpninfo->dtls_ssl),
 			 BIO_CTRL_DGRAM_SET_NEXT_TIMEOUT, 0,
 			 &(vpninfo->dtls_ssl->d1->next_timeout));
-#elif defined (BIO_CTRL_DGRAM_SET_TIMEOUT)
+#elif defined(BIO_CTRL_DGRAM_SET_TIMEOUT)
 		/*
 		 * OK, here it gets more fun... this shoul handle the case
 		 * of older OpenSSL which has the Cisco DTLS compatibility
@@ -298,9 +299,19 @@ int dtls_try_handshake(struct openconnect_info *vpninfo)
 
 	ret = SSL_get_error(vpninfo->new_dtls_ssl, ret);
 	if (ret == SSL_ERROR_WANT_WRITE || ret == SSL_ERROR_WANT_READ) {
+		static int badossl_bitched = 0;
 		if (time(NULL) < vpninfo->new_dtls_started + 5)
 			return 0;
-		vpn_progress(vpninfo, PRG_TRACE, _("DTLS handshake timed out\n"));
+		if (((OPENSSL_VERSION_NUMBER >= 0x100000b0L && OPENSSL_VERSION_NUMBER <= 0x100000c0L) || \
+		     (OPENSSL_VERSION_NUMBER >= 0x10001040L && OPENSSL_VERSION_NUMBER <= 0x10001060L) || \
+		     OPENSSL_VERSION_NUMBER == 0x10002000L) && !badossl_bitched) {
+			badossl_bitched = 1;
+			vpn_progress(vpninfo, PRG_ERR, _("DTLS handshake timed out\n"));
+			vpn_progress(vpninfo, PRG_ERR, _("This is probably because your OpenSSL is broken\n"
+				"See http://rt.openssl.org/Ticket/Display.html?id=2984\n"));
+		} else {
+			vpn_progress(vpninfo, PRG_TRACE, _("DTLS handshake timed out\n"));
+		}
 	}
 
 	vpn_progress(vpninfo, PRG_ERR, _("DTLS handshake failed: %d\n"), ret);
@@ -331,7 +342,7 @@ int dtls_try_handshake(struct openconnect_info *vpninfo)
 	return -EINVAL;
 }
 
-#elif defined (DTLS_GNUTLS)
+#elif defined(DTLS_GNUTLS)
 #include <gnutls/dtls.h>
 
 struct {
@@ -342,6 +353,8 @@ struct {
 } gnutls_dtls_ciphers[] = {
 	{ "AES128-SHA", GNUTLS_CIPHER_AES_128_CBC, GNUTLS_MAC_SHA1,
 	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-128-CBC:+SHA1:+RSA:%COMPAT:%DISABLE_SAFE_RENEGOTIATION" },
+	{ "AES256-SHA", GNUTLS_CIPHER_AES_256_CBC, GNUTLS_MAC_SHA1,
+	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-256-CBC:+SHA1:+RSA:%COMPAT:%DISABLE_SAFE_RENEGOTIATION" },
 	{ "DES-CBC3-SHA", GNUTLS_CIPHER_3DES_CBC, GNUTLS_MAC_SHA1,
 	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+3DES-CBC:+SHA1:+RSA:%COMPAT:%DISABLE_SAFE_RENEGOTIATION" },
 };
@@ -544,11 +557,13 @@ int connect_dtls_socket(struct openconnect_info *vpninfo)
 				     _("Unknown protocol family %d. Cannot do DTLS\n"),
 				     vpninfo->peer_addr->sa_family);
 			vpninfo->dtls_attempt_period = 0;
+			close(dtls_fd);
 			return -EINVAL;
 		}
 
 		if (bind(dtls_fd, (struct sockaddr *)&dtls_bind_addr, dtls_bind_addrlen)) {
 			perror(_("Bind UDP socket for DTLS"));
+			close(dtls_fd);
 			return -EINVAL;
 		}
 	}
@@ -583,9 +598,9 @@ int connect_dtls_socket(struct openconnect_info *vpninfo)
 static int dtls_restart(struct openconnect_info *vpninfo)
 {
 	if (vpninfo->dtls_ssl) {
-#if defined (DTLS_OPENSSL)
+#if defined(DTLS_OPENSSL)
 		SSL_free(vpninfo->dtls_ssl);
-#elif defined (DTLS_GNUTLS)
+#elif defined(DTLS_GNUTLS)
 		gnutls_deinit(vpninfo->dtls_ssl);
 #endif
 		close(vpninfo->dtls_fd);
@@ -605,13 +620,13 @@ int setup_dtls(struct openconnect_info *vpninfo)
 	struct vpn_option *dtls_opt = vpninfo->dtls_options;
 	int dtls_port = 0;
 
-#if defined (OPENCONNECT_GNUTLS) && defined (DTLS_OPENSSL)
+#if defined(OPENCONNECT_GNUTLS) && defined(DTLS_OPENSSL)
 	/* If we're using GnuTLS for authentication but OpenSSL for DTLS,
 	   we'll need to initialise OpenSSL now... */
-	SSL_library_init ();
-	ERR_clear_error ();
-	SSL_load_error_strings ();
-	OpenSSL_add_all_algorithms ();
+	SSL_library_init();
+	ERR_clear_error();
+	SSL_load_error_strings();
+	OpenSSL_add_all_algorithms();
 #endif
 
 	while (dtls_opt) {
@@ -665,7 +680,7 @@ int setup_dtls(struct openconnect_info *vpninfo)
 		return -EINVAL;
 
 	vpn_progress(vpninfo, PRG_TRACE,
-		     _("DTLS connected. DPD %d, Keepalive %d\n"),
+		     _("DTLS initialised. DPD %d, Keepalive %d\n"),
 		     vpninfo->dtls_times.dpd, vpninfo->dtls_times.keepalive);
 
 	return 0;
@@ -701,7 +716,7 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 
 		vpninfo->dtls_times.last_rx = time(NULL);
 
-		switch(buf[0]) {
+		switch (buf[0]) {
 		case AC_PKT_DATA:
 			dtls_pkt->len = len - 1;
 			queue_packet(&vpninfo->incoming_queue, dtls_pkt);
@@ -834,7 +849,7 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 			}
 			return 1;
 		}
-#elif defined (DTLS_GNUTLS)
+#elif defined(DTLS_GNUTLS)
 		ret = gnutls_record_send(vpninfo->dtls_ssl, &this->hdr[7], this->len + 1);
 		if (ret <= 0) {
 			if (ret != GNUTLS_E_AGAIN) {
@@ -859,11 +874,10 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 }
 #else /* !HAVE_DTLS */
 #warning Your SSL library does not seem to support Cisco DTLS compatibility
- int setup_dtls(struct openconnect_info *vpninfo)
+int setup_dtls(struct openconnect_info *vpninfo)
 {
 	vpn_progress(vpninfo, PRG_ERR,
 		     _("Built against SSL library with no Cisco DTLS support\n"));
 	return -EINVAL;
 }
 #endif
-
