@@ -79,7 +79,7 @@ int openconnect_SSL_write(struct openconnect_info *vpninfo, char *buf, size_t le
 		if (done > 0)
 			len -= done;
 		else if (done != GNUTLS_E_AGAIN) {
-			vpn_progress(vpninfo, PRG_ERR, _("Failed to write to SSL socket: %s"),
+			vpn_progress(vpninfo, PRG_ERR, _("Failed to write to SSL socket: %s\n"),
 				     gnutls_strerror(done));
 			return -EIO;
 		} else {
@@ -88,7 +88,7 @@ int openconnect_SSL_write(struct openconnect_info *vpninfo, char *buf, size_t le
 
 			FD_ZERO(&wr_set);
 			FD_ZERO(&rd_set);
-			
+
 			if (gnutls_record_get_direction(vpninfo->https_sess))
 				FD_SET(vpninfo->ssl_fd, &wr_set);
 			else
@@ -119,13 +119,13 @@ int openconnect_SSL_read(struct openconnect_info *vpninfo, char *buf, size_t len
 		int maxfd = vpninfo->ssl_fd;
 
 		if (done != GNUTLS_E_AGAIN) {
-			vpn_progress(vpninfo, PRG_ERR, _("Failed to read from SSL socket: %s"),
+			vpn_progress(vpninfo, PRG_ERR, _("Failed to read from SSL socket: %s\n"),
 				     gnutls_strerror(done));
 			return -EIO;
 		} else {
 			FD_ZERO(&wr_set);
 			FD_ZERO(&rd_set);
-			
+
 			if (gnutls_record_get_direction(vpninfo->https_sess))
 				FD_SET(vpninfo->ssl_fd, &wr_set);
 			else
@@ -180,10 +180,10 @@ int openconnect_SSL_gets(struct openconnect_info *vpninfo, char *buf, size_t len
 		} else {
 			fd_set rd_set, wr_set;
 			int maxfd = vpninfo->ssl_fd;
-			
+
 			FD_ZERO(&rd_set);
 			FD_ZERO(&wr_set);
-			
+
 			if (gnutls_record_get_direction(vpninfo->https_sess))
 				FD_SET(vpninfo->ssl_fd, &wr_set);
 			else
@@ -434,7 +434,7 @@ static int count_x509_certificates(gnutls_datum_t *datum)
 		p += 11;
 		if (!strncmp(p, "CERTIFICATE", 11) ||
 		    !strncmp(p, "X509 CERTIFICATE", 16))
-		    count++;
+			count++;
 	}
 	return count;
 }
@@ -451,7 +451,7 @@ static int get_cert_name(gnutls_x509_crt_t cert, char *name, size_t namelen)
 	return 0;
 }
 
-#if defined (HAVE_P11KIT) || defined (HAVE_TROUSERS)
+#if defined(HAVE_P11KIT) || defined(HAVE_TROUSERS)
 #ifndef HAVE_GNUTLS_CERTIFICATE_SET_KEY
 /* For GnuTLS 2.12 even if we *have* a privkey (as we do for PKCS#11), we
    can't register it. So we have to use the cert_callback function. This
@@ -555,7 +555,7 @@ static int assign_privkey(struct openconnect_info *vpninfo,
 	if (!pcerts)
 		return GNUTLS_E_MEMORY_ERROR;
 
-	for (i=0 ; i < nr_certs; i++) {
+	for (i = 0 ; i < nr_certs; i++) {
 		err = gnutls_pcert_import_x509(pcerts + i, certs[i], 0);
 		if (err) {
 			vpn_progress(vpninfo, PRG_ERR,
@@ -572,13 +572,29 @@ static int assign_privkey(struct openconnect_info *vpninfo,
 			     _("Setting PKCS#11 certificate failed: %s\n"),
 			     gnutls_strerror(err));
 	free_pcerts:
-		for (i=0 ; i < nr_certs; i++)
+		for (i = 0 ; i < nr_certs; i++)
 			gnutls_pcert_deinit(pcerts + i);
-		free (pcerts);
+		free(pcerts);
 	}
 	return err;
 }
 #endif /* !SET_KEY */
+
+static int verify_signed_data(gnutls_pubkey_t pubkey, gnutls_privkey_t privkey,
+			      const gnutls_datum_t *data, const gnutls_datum_t *sig)
+{
+#ifdef HAVE_GNUTLS_PK_TO_SIGN
+	gnutls_sign_algorithm_t algo = GNUTLS_SIGN_RSA_SHA1; /* TPM keys */
+
+	if (privkey != OPENCONNECT_TPM_PKEY)
+		algo = gnutls_pk_to_sign(gnutls_privkey_get_pk_algorithm(privkey, NULL),
+					 GNUTLS_DIG_SHA1);
+
+	return gnutls_pubkey_verify_data2(pubkey, algo, 0, data, sig);
+#else
+	return gnutls_pubkey_verify_data(pubkey, 0, data, sig);
+#endif
+}
 #endif /* (P11KIT || TROUSERS) */
 
 static int openssl_hash_password(struct openconnect_info *vpninfo, char *pass,
@@ -652,7 +668,7 @@ static int import_openssl_pem(struct openconnect_info *vpninfo,
 	else if (type == 'R')
 		begin = "RSA PRIVATE KEY";
 	else if (type == 'D')
-		 begin = "DSA PRIVATE KEY";
+		begin = "DSA PRIVATE KEY";
 	else
 		return -EINVAL;
 
@@ -868,7 +884,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 {
 	gnutls_datum_t fdata;
 	gnutls_x509_privkey_t key = NULL;
-#if defined (HAVE_P11KIT) || defined (HAVE_TROUSERS)
+#if defined(HAVE_P11KIT) || defined(HAVE_TROUSERS)
 	gnutls_privkey_t pkey = NULL;
 	gnutls_datum_t pkey_sig = {NULL, 0};
 	void *dummy_hash_data = &load_certificate;
@@ -1040,7 +1056,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
  got_certs:
 	/* Now we have either a single certificate in 'cert', or an array of
 	   them in extra_certs[]. Next we look for the private key ... */
-#if defined (HAVE_P11KIT)
+#if defined(HAVE_P11KIT)
 	if (key_is_p11) {
 		vpn_progress(vpninfo, PRG_TRACE,
 			     _("Using PKCS#11 key %s\n"), key_url);
@@ -1107,14 +1123,14 @@ static int load_certificate(struct openconnect_info *vpninfo)
 				s = sizeof(token->model) + 1;
 				if (!gnutls_pkcs11_obj_get_info(crt, GNUTLS_PKCS11_OBJ_TOKEN_MODEL,
 								buf, &s)) {
-					    s--;
-					    memcpy(token->model, buf, s);
-					    memset(token->model + s, ' ',
-						   sizeof(token->model) - s);
+					s--;
+					memcpy(token->model, buf, s);
+					memset(token->model + s, ' ',
+					       sizeof(token->model) - s);
 				}
 			}
 			if (!token->serialNumber[0]) {
-				s = sizeof(token->serialNumber + 1);
+				s = sizeof(token->serialNumber) + 1;
 				if (!gnutls_pkcs11_obj_get_info(crt, GNUTLS_PKCS11_OBJ_TOKEN_SERIAL,
 								buf, &s)) {
 					s--;
@@ -1306,11 +1322,11 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		goto out;
 	}
 	/* If extra_certs[] is NULL, we have one candidate in 'cert' to check. */
-	for (i = 0; i < (extra_certs?nr_extra_certs:1); i++) {
+	for (i = 0; i < (extra_certs ? nr_extra_certs : 1); i++) {
 		unsigned char cert_id[20];
 		size_t cert_id_size = sizeof(cert_id);
 
-		err = gnutls_x509_crt_get_key_id(extra_certs?extra_certs[i]:cert, 0, cert_id, &cert_id_size);
+		err = gnutls_x509_crt_get_key_id(extra_certs ? extra_certs[i] : cert, 0, cert_id, &cert_id_size);
 		if (err)
 			continue;
 
@@ -1326,7 +1342,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	   enabled we'll fall straight through the bit at match_cert: below, and go
 	   directly to the bit where it prints the 'no match found' error and exits. */
 
-#if defined (HAVE_P11KIT) || defined (HAVE_TROUSERS)
+#if defined(HAVE_P11KIT) || defined(HAVE_TROUSERS)
  match_cert:
 	/* If we have a privkey from PKCS#11 or TPM, we can't do the simple comparison
 	   of key ID that we do for software keys to find which certificate is a
@@ -1353,11 +1369,11 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		}
 
 		/* If extra_certs[] is NULL, we have one candidate in 'cert' to check. */
-		for (i=0; i < (extra_certs?nr_extra_certs:1); i++) {
+		for (i = 0; i < (extra_certs ? nr_extra_certs : 1); i++) {
 			gnutls_pubkey_t pubkey;
 
 			gnutls_pubkey_init(&pubkey);
-			err = gnutls_pubkey_import_x509(pubkey, extra_certs?extra_certs[i]:cert, 0);
+			err = gnutls_pubkey_import_x509(pubkey, extra_certs ? extra_certs[i] : cert, 0);
 			if (err) {
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("Error validating signature against certificate: %s\n"),
@@ -1366,7 +1382,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 				gnutls_pubkey_deinit(pubkey);
 				continue;
 			}
-			err = gnutls_pubkey_verify_data(pubkey, 0, &fdata, &pkey_sig);
+			err = verify_signed_data(pubkey, pkey, &fdata, &pkey_sig);
 			gnutls_pubkey_deinit(pubkey);
 
 			if (err >= 0) {
@@ -1428,6 +1444,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	}
 	while (1) {
 		gnutls_x509_crt_t issuer;
+		void *tmp;
 
 		for (i = 0; i < nr_extra_certs; i++) {
 			if (extra_certs[i] &&
@@ -1444,7 +1461,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 			err = gnutls_certificate_get_issuer(vpninfo->https_cred,
 							    last_cert, &issuer, 0);
 			if (err)
- 				break;
+				break;
 
 			/* The check_issuer_sanity() function works fine as a workaround where
 			   it was used above, but when gnutls_certificate_get_issuer() returns
@@ -1470,9 +1487,11 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		}
 
 		/* OK, we found a new cert to add to our chain. */
+		tmp = supporting_certs;
 		supporting_certs = gnutls_realloc(supporting_certs,
 						  sizeof(cert) * ++nr_supporting_certs);
 		if (!supporting_certs) {
+			gnutls_free(tmp);
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Failed to allocate memory for supporting certificates\n"));
 			/* The world is probably about to end, but try without them anyway */
@@ -1503,7 +1522,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	   key and certs. GnuTLS makes us do this differently for X509 privkeys
 	   vs. TPM/PKCS#11 "generic" privkeys, and the latter is particularly
 	   'fun' for GnuTLS 2.12... */
-#if defined (HAVE_P11KIT) || defined (HAVE_TROUSERS)
+#if defined(HAVE_P11KIT) || defined(HAVE_TROUSERS)
 	if (pkey) {
 		err = assign_privkey(vpninfo, pkey,
 				     supporting_certs ? supporting_certs : &cert,
@@ -1548,7 +1567,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	}
 	gnutls_free(extra_certs);
 
-#if defined (HAVE_P11KIT) || defined (HAVE_TROUSERS)
+#if defined(HAVE_P11KIT) || defined(HAVE_TROUSERS)
 	if (pkey && pkey != OPENCONNECT_TPM_PKEY)
 		gnutls_privkey_deinit(pkey);
 	/* If we support arbitrary privkeys, we might have abused fdata.data
@@ -1578,7 +1597,7 @@ static int get_cert_fingerprint(struct openconnect_info *vpninfo,
 	if (gnutls_x509_crt_get_fingerprint(cert, algo, md, &md_size))
 		return -EIO;
 
-	for (i=0; i < md_size; i++)
+	for (i = 0; i < md_size; i++)
 		sprintf(&buf[i*2], "%02X", md[i]);
 
 	return 0;
@@ -1604,11 +1623,11 @@ char *openconnect_get_cert_details(struct openconnect_info *vpninfo,
 
 	if (gnutls_x509_crt_print(cert, GNUTLS_CRT_PRINT_FULL, &buf))
 		return NULL;
-	
+
 	/* Just in case gnutls_free() isn't free(), we can't steal it. */
 	ret = strdup((char *)buf.data);
 	gnutls_free(buf.data);
-	
+
 	return ret;
 }
 
@@ -1618,7 +1637,7 @@ int openconnect_get_cert_DER(struct openconnect_info *vpninfo,
 	size_t l = 0;
 	unsigned char *ret = NULL;
 
-	if (gnutls_x509_crt_export(cert, GNUTLS_X509_FMT_DER, ret, &l) != 
+	if (gnutls_x509_crt_export(cert, GNUTLS_X509_FMT_DER, ret, &l) !=
 	    GNUTLS_E_SHORT_MEMORY_BUFFER)
 		return -EIO;
 
@@ -1628,7 +1647,7 @@ int openconnect_get_cert_DER(struct openconnect_info *vpninfo,
 
 	if (gnutls_x509_crt_export(cert, GNUTLS_X509_FMT_DER, ret, &l)) {
 		free(ret);
- 		return -EIO;
+		return -EIO;
 	}
 	*buf = ret;
 	return l;
@@ -1648,7 +1667,7 @@ static int verify_peer(gnutls_session_t session)
 		vpninfo->peer_cert = NULL;
 	}
 
-	cert_list = gnutls_certificate_get_peers (session, &cert_list_size);
+	cert_list = gnutls_certificate_get_peers(session, &cert_list_size);
 	if (!cert_list) {
 		vpn_progress(vpninfo, PRG_ERR, _("Server presented no certificate\n"));
 		return GNUTLS_E_CERTIFICATE_ERROR;
@@ -1658,14 +1677,14 @@ static int verify_peer(gnutls_session_t session)
 		unsigned char sha1bin[SHA1_SIZE];
 		char fingerprint[(SHA1_SIZE * 2) + 1];
 		int i;
-		
+
 		err = openconnect_sha1(sha1bin, cert_list[0].data, cert_list[0].size);
 		if (err) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Could not calculate SHA1 of server's certificate\n"));
 			return GNUTLS_E_CERTIFICATE_ERROR;
 		}
-		for (i=0; i < SHA1_SIZE; i++)
+		for (i = 0; i < SHA1_SIZE; i++)
 			sprintf(&fingerprint[i*2], "%02X", sha1bin[i]);
 
 		if (strcasecmp(vpninfo->servercert, fingerprint)) {
@@ -1676,7 +1695,7 @@ static int verify_peer(gnutls_session_t session)
 		return 0;
 	}
 
-	err = gnutls_certificate_verify_peers2 (session, &status);
+	err = gnutls_certificate_verify_peers2(session, &status);
 	if (err) {
 		vpn_progress(vpninfo, PRG_ERR, _("Error checking server cert status\n"));
 		return GNUTLS_E_CERTIFICATE_ERROR;
@@ -1754,8 +1773,8 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 						       DEFAULT_SYSTEM_CAFILE,
 						       GNUTLS_X509_FMT_PEM);
 #endif
-		gnutls_certificate_set_verify_function (vpninfo->https_cred,
-							verify_peer);
+		gnutls_certificate_set_verify_function(vpninfo->https_cred,
+						       verify_peer);
 
 #ifdef ANDROID_KEYSTORE
 		if (vpninfo->cafile && !strncmp(vpninfo->cafile, "keystore:", 9)) {
@@ -1763,13 +1782,16 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 			unsigned int nr_certs;
 
 			err = load_datum(vpninfo, &datum, vpninfo->cafile);
-			if (err < 0)
+			if (err < 0) {
+				gnutls_certificate_free_credentials(vpninfo->https_cred);
+				vpninfo->https_cred = NULL;
 				return err;
+			}
 
 			/* For GnuTLS 3.x We should use gnutls_x509_crt_list_import2() */
 			nr_certs = count_x509_certificates(&datum);
 			if (nr_certs) {
-				gnutls_x509_crt *certs;
+				gnutls_x509_crt_t *certs;
 				int i;
 
 				certs = calloc(nr_certs, sizeof(*certs));
@@ -1777,6 +1799,8 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 					vpn_progress(vpninfo, PRG_ERR,
 						     _("Failed to allocate memory for cafile certs\n"));
 					gnutls_free(datum.data);
+					gnutls_certificate_free_credentials(vpninfo->https_cred);
+					vpninfo->https_cred = NULL;
 					close(ssl_sock);
 					return -ENOMEM;
 				}
@@ -1796,6 +1820,8 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 					vpn_progress(vpninfo, PRG_ERR,
 						     _("Failed to read certs from cafile: %s\n"),
 						     gnutls_strerror(err));
+					gnutls_certificate_free_credentials(vpninfo->https_cred);
+					vpninfo->https_cred = NULL;
 					close(ssl_sock);
 					return -EINVAL;
 				}
@@ -1810,6 +1836,8 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("Failed to open CA file '%s': %s\n"),
 					     vpninfo->cafile, gnutls_strerror(err));
+				gnutls_certificate_free_credentials(vpninfo->https_cred);
+				vpninfo->https_cred = NULL;
 				close(ssl_sock);
 				return -EINVAL;
 			}
@@ -1820,40 +1848,45 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 			if (err) {
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("Loading certificate failed. Aborting.\n"));
+				gnutls_certificate_free_credentials(vpninfo->https_cred);
+				vpninfo->https_cred = NULL;
 				close(ssl_sock);
 				return err;
 			}
 		}
 	}
-	gnutls_init (&vpninfo->https_sess, GNUTLS_CLIENT);
-	gnutls_session_set_ptr (vpninfo->https_sess, (void *) vpninfo);
+	gnutls_init(&vpninfo->https_sess, GNUTLS_CLIENT);
+	gnutls_session_set_ptr(vpninfo->https_sess, (void *) vpninfo);
 #if defined(HAVE_TROUSERS) && !defined(HAVE_GNUTLS_CERTIFICATE_SET_KEY)
 	if (vpninfo->my_pkey == OPENCONNECT_TPM_PKEY)
 		gnutls_sign_callback_set(vpninfo->https_sess, gtls2_tpm_sign_cb, vpninfo);
 #endif
 
-	err = gnutls_priority_set_direct (vpninfo->https_sess,
-					  "NORMAL:-VERS-TLS-ALL:+VERS-TLS1.0:"
+	err = gnutls_priority_set_direct(vpninfo->https_sess,
+					 "NORMAL:-VERS-TLS-ALL:+VERS-TLS1.0:"
 #if GNUTLS_VERSION_MAJOR >= 3
-					  "-CURVE-ALL:"
+					 "-CURVE-ALL:"
 #endif
-					  "%COMPAT:%DISABLE_SAFE_RENEGOTIATION:%LATEST_RECORD_VERSION",
-					  NULL);
+					 "%COMPAT:%DISABLE_SAFE_RENEGOTIATION:%LATEST_RECORD_VERSION",
+					 NULL);
 	if (err) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Failed to set TLS priority string: %s\n"),
 			     gnutls_strerror(err));
+		gnutls_deinit(vpninfo->https_sess);
+		vpninfo->https_sess = NULL;
+		close(ssl_sock);
 		return -EIO;
 	}
 
-	gnutls_record_disable_padding (vpninfo->https_sess);
-	gnutls_credentials_set (vpninfo->https_sess, GNUTLS_CRD_CERTIFICATE, vpninfo->https_cred);
+	gnutls_record_disable_padding(vpninfo->https_sess);
+	gnutls_credentials_set(vpninfo->https_sess, GNUTLS_CRD_CERTIFICATE, vpninfo->https_cred);
 	gnutls_transport_set_ptr(vpninfo->https_sess, /* really? */(gnutls_transport_ptr_t)(long) ssl_sock);
 
 	vpn_progress(vpninfo, PRG_INFO, _("SSL negotiation with %s\n"),
 		     vpninfo->hostname);
 
-	while ((err = gnutls_handshake (vpninfo->https_sess))) {
+	while ((err = gnutls_handshake(vpninfo->https_sess))) {
 		if (err == GNUTLS_E_AGAIN) {
 			fd_set rd_set, wr_set;
 			int maxfd = ssl_sock;
