@@ -3,6 +3,7 @@
  *
  * Copyright © 2008-2012 Intel Corporation.
  * Copyright © 2008 Nick Andrew <nick@nick-andrew.net>
+ * Copyright © 2013 John Morrissey <jwm@horde.net>
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -29,7 +30,7 @@
 #endif
 
 #include <stdio.h>
-#ifdef ANDROID
+#ifdef __ANDROID__
 #include <android/log.h>
 #else
 #include <syslog.h>
@@ -66,11 +67,13 @@ static int validate_peer_cert(void *_vpninfo,
 			      const char *reason);
 static int process_auth_form(void *_vpninfo,
 			     struct oc_auth_form *form);
+static void init_token(struct openconnect_info *vpninfo,
+		       oc_token_mode_t token_mode, const char *token_str);
 
 /* A sanity check that the openconnect executable is running against a
    library of the same version */
 #define openconnect_version_str openconnect_binary_version
-#include "version.c"
+#include <version.c>
 #undef openconnect_version_str
 
 int verbose = PRG_INFO;
@@ -92,6 +95,7 @@ enum {
 	OPT_CSD_WRAPPER,
 	OPT_DISABLE_IPV6,
 	OPT_DTLS_CIPHERS,
+	OPT_DUMP_HTTP,
 	OPT_FORCE_DPD,
 	OPT_KEY_PASSWORD_FROM_FSID,
 	OPT_LIBPROXY,
@@ -100,6 +104,7 @@ enum {
 	OPT_NO_HTTP_KEEPALIVE,
 	OPT_NO_PASSWD,
 	OPT_NO_PROXY,
+	OPT_NO_XMLPOST,
 	OPT_PIDFILE,
 	OPT_PASSWORD_ON_STDIN,
 	OPT_PRINTCOOKIE,
@@ -108,6 +113,9 @@ enum {
 	OPT_USERAGENT,
 	OPT_NON_INTER,
 	OPT_DTLS_LOCAL_PORT,
+	OPT_TOKEN_MODE,
+	OPT_TOKEN_SECRET,
+	OPT_OS,
 };
 
 #ifdef __sun__
@@ -171,6 +179,11 @@ static struct option long_options[] = {
 	OPTION("force-dpd", 1, OPT_FORCE_DPD),
 	OPTION("non-inter", 0, OPT_NON_INTER),
 	OPTION("dtls-local-port", 1, OPT_DTLS_LOCAL_PORT),
+	OPTION("token-mode", 1, OPT_TOKEN_MODE),
+	OPTION("token-secret", 1, OPT_TOKEN_SECRET),
+	OPTION("os", 1, OPT_OS),
+	OPTION("no-xmlpost", 0, OPT_NO_XMLPOST),
+	OPTION("dump-http-traffic", 0, OPT_DUMP_HTTP),
 	OPTION(NULL, 0, 0)
 };
 
@@ -184,9 +197,9 @@ static void print_build_opts(void)
 {
 	const char *comma = ", ", *sep = comma + 1;
 
-#if defined (OPENCONNECT_OPENSSL)
+#if defined(OPENCONNECT_OPENSSL)
 	printf(_("Using OpenSSL. Features present:"));
-#elif defined (OPENCONNECT_GNUTLS)
+#elif defined(OPENCONNECT_GNUTLS)
 	printf(_("Using GnuTLS. Features present:"));
 #endif
 
@@ -194,7 +207,7 @@ static void print_build_opts(void)
 		printf("%sTPM", sep);
 		sep = comma;
 	}
-#if defined (OPENCONNECT_OPENSSL) && defined (HAVE_ENGINE)
+#if defined(OPENCONNECT_OPENSSL) && defined(HAVE_ENGINE)
 	else {
 		printf("%sTPM (%s)", sep, _("OpenSSL ENGINE not present"));
 		sep = comma;
@@ -204,10 +217,18 @@ static void print_build_opts(void)
 		printf("%sPKCS#11", sep);
 		sep = comma;
 	}
+	if (openconnect_has_stoken_support()) {
+		printf("%sRSA software token", sep);
+		sep = comma;
+	}
+	if (openconnect_has_oath_support()) {
+		printf("%sTOTP software token", sep);
+		sep = comma;
+	}
 
 #ifdef HAVE_DTLS
 	printf("%sDTLS", sep);
-#if defined (OPENCONNECT_GNUTLS) && defined (DTLS_OPENSSL)
+#if defined(OPENCONNECT_GNUTLS) && defined(DTLS_OPENSSL)
 	printf(" (%s)", _("using OpenSSL"));
 #endif
 	printf("\n");
@@ -257,6 +278,7 @@ static void usage(void)
 	printf("  -u, --user=NAME                 %s\n", _("Set login username"));
 	printf("  -V, --version                   %s\n", _("Report version number"));
 	printf("  -v, --verbose                   %s\n", _("More output"));
+	printf("      --dump-http-traffic         %s\n", _("Dump HTTP authentication traffic (implies --verbose"));
 	printf("  -x, --xmlconfig=CONFIG          %s\n", _("XML config file"));
 	printf("      --authgroup=GROUP           %s\n", _("Choose authentication login selection"));
 	printf("      --authenticate              %s\n", _("Authenticate only and print login info"));
@@ -269,11 +291,21 @@ static void usage(void)
 	printf("      --no-http-keepalive         %s\n", _("Disable HTTP connection re-use"));
 	printf("      --no-passwd                 %s\n", _("Disable password/SecurID authentication"));
 	printf("      --no-cert-check             %s\n", _("Do not require server SSL cert to be valid"));
+	printf("      --no-xmlpost                %s\n", _("Do not attempt XML POST authentication"));
 	printf("      --non-inter                 %s\n", _("Do not expect user input; exit if it is required"));
 	printf("      --passwd-on-stdin           %s\n", _("Read password from standard input"));
+	printf("      --token-mode=MODE           %s\n", _("Software token type: rsa or totp"));
+	printf("      --token-secret=STRING       %s\n", _("Software token secret"));
+#ifndef HAVE_LIBSTOKEN
+	printf("                                  %s\n", _("(NOTE: libstoken (RSA SecurID) disabled in this build)"));
+#endif
+#ifndef HAVE_LIBOATH
+	printf("                                  %s\n", _("(NOTE: liboath (TOTP) disabled in this build)"));
+#endif
 	printf("      --reconnect-timeout         %s\n", _("Connection retry timeout in seconds"));
 	printf("      --servercert=FINGERPRINT    %s\n", _("Server's certificate SHA1 fingerprint"));
 	printf("      --useragent=STRING          %s\n", _("HTTP header User-Agent: field"));
+	printf("      --os=STRING                 %s\n", _("OS type (linux,linux-64,mac,win) to report"));
 	printf("      --dtls-local-port=PORT      %s\n", _("Set local port for DTLS datagrams"));
 	printf("\n");
 
@@ -314,14 +346,14 @@ static int config_line_num = 0;
  *
  * 1. We only care about it transiently and it can be lost entirely
  *    (e.g. vpninfo->reconnect_timeout = atoi(config_arg);
- * 2. We need to kep it, but it's a static string and will never be freed
+ * 2. We need to keep it, but it's a static string and will never be freed
  *    so when it's part of argv[] we can use it in place, but when it comes
  *    from a file we have to strdup() because otherwise it'll be overwritten.
  *    For this we use the keep_config_arg() macro below.
  * 3. It may be freed during normal operation, so we have to use strdup()
  *    even when it's an option from argv[]. (e.g. vpninfo->cert_password).
  */
-#define keep_config_arg() (config_file?strdup(config_arg):config_arg)
+#define keep_config_arg() (config_file && config_arg ? strdup(config_arg) : config_arg)
 
 static int next_option(int argc, char **argv, char **config_arg)
 {
@@ -330,7 +362,7 @@ static int next_option(int argc, char **argv, char **config_arg)
 	static size_t line_size = 0;
 
 	ssize_t llen;
-	int opt, optlen;
+	int opt, optlen = 0;
 	struct option *this;
 	char *line;
 	int ate_equals = 0;
@@ -338,7 +370,7 @@ static int next_option(int argc, char **argv, char **config_arg)
  next:
 	if (!config_file) {
 		opt = getopt_long(argc, argv,
-				  "bC:c:e:Ddg:hi:k:lp:P:Q:qSs:U:u:Vvx:",
+				  "bC:c:Dde:g:hi:k:lm:P:p:Q:qSs:U:u:Vvx:",
 				  long_options, NULL);
 
 		*config_arg = optarg;
@@ -383,7 +415,7 @@ static int next_option(int argc, char **argv, char **config_arg)
 		if (!strncmp(this->name, line, optlen) &&
 		    (!line[optlen] || line[optlen] == ' ' || line[optlen] == '\t' ||
 		     line[optlen] == '='))
-		    break;
+			break;
 	}
 	if (!this->name) {
 		char *l;
@@ -405,10 +437,12 @@ static int next_option(int argc, char **argv, char **config_arg)
 		fprintf(stderr, _("Option '%s' does not take an argument at line %d\n"),
 			this->name, config_line_num);
 		return '?';
-	} else if (this->has_arg && !*line) {
+	} else if (this->has_arg == 1 && !*line) {
 		fprintf(stderr, _("Option '%s' requires an argument at line %d\n"),
 			this->name, config_line_num);
 		return '?';
+	} else if (this->has_arg == 2 && !*line) {
+		line = NULL;
 	}
 
 	config_line_num++;
@@ -431,6 +465,8 @@ int main(int argc, char **argv)
 	char *pidfile = NULL;
 	FILE *fp = NULL;
 	char *config_arg;
+	char *token_str = NULL;
+	oc_token_mode_t token_mode = OC_TOKEN_MODE_NONE;
 
 #ifdef ENABLE_NLS
 	bindtextdomain("openconnect", LOCALEDIR);
@@ -442,7 +478,7 @@ int main(int argc, char **argv)
 				  "         the libopenconnect library is %s\n"),
 			openconnect_binary_version, openconnect_version_str);
 	}
-			
+
 	openconnect_init_ssl();
 
 	vpninfo = malloc(sizeof(*vpninfo));
@@ -463,11 +499,7 @@ int main(int argc, char **argv)
 	vpninfo->reconnect_timeout = 300;
 	vpninfo->uid_csd = 0;
 	/* We could let them override this on the command line some day, perhaps */
-#ifdef __APPLE__
-	vpninfo->csd_xmltag = "csdMac";
-#else
-	vpninfo->csd_xmltag = "csdLinux";
-#endif
+	openconnect_set_reported_os(vpninfo, NULL);
 	vpninfo->uid_csd = 0;
 	vpninfo->uid_csd_given = 0;
 	vpninfo->validate_peer_cert = validate_peer_cert;
@@ -476,6 +508,7 @@ int main(int argc, char **argv)
 	vpninfo->cert_expire_warning = 60 * 86400;
 	vpninfo->vpnc_script = DEFAULT_VPNCSCRIPT;
 	vpninfo->cancel_fd = -1;
+	vpninfo->xmlpost = 1;
 
 	if (!uname(&utsbuf))
 		vpninfo->localname = utsbuf.nodename;
@@ -526,9 +559,8 @@ int main(int argc, char **argv)
 		case OPT_COOKIE_ON_STDIN:
 			read_stdin(&vpninfo->cookie);
 			/* If the cookie is empty, ignore it */
-			if (! *vpninfo->cookie) {
+			if (!*vpninfo->cookie)
 				vpninfo->cookie = NULL;
-			}
 			break;
 		case OPT_PASSWORD_ON_STDIN:
 			read_stdin(&vpninfo->password);
@@ -536,6 +568,9 @@ int main(int argc, char **argv)
 		case OPT_NO_PASSWD:
 			vpninfo->nopasswd = 1;
 			break;
+		case OPT_NO_XMLPOST:
+			vpninfo->xmlpost = 0;
+			break;
 		case OPT_NON_INTER:
 			non_inter = 1;
 			break;
@@ -598,7 +633,7 @@ int main(int argc, char **argv)
 		case 'p':
 			vpninfo->cert_password = strdup(config_arg);
 			break;
-		case 'P': 
+		case 'P':
 			proxy = keep_config_arg();
 			autoproxy = 0;
 			break;
@@ -673,6 +708,8 @@ int main(int argc, char **argv)
 		case 'q':
 			verbose = PRG_ERR;
 			break;
+		case OPT_DUMP_HTTP:
+			vpninfo->dump_http_traffic = 1;
 		case 'v':
 			verbose = PRG_TRACE;
 			break;
@@ -697,6 +734,27 @@ int main(int argc, char **argv)
 		case OPT_DTLS_LOCAL_PORT:
 			vpninfo->dtls_local_port = atoi(config_arg);
 			break;
+		case OPT_TOKEN_MODE:
+			if (strcasecmp(config_arg, "rsa") == 0) {
+				token_mode = OC_TOKEN_MODE_STOKEN;
+			} else if (strcasecmp(config_arg, "totp") == 0) {
+				token_mode = OC_TOKEN_MODE_TOTP;
+			} else {
+				fprintf(stderr, _("Invalid software token mode \"%s\"\n"),
+					config_arg);
+				exit(1);
+			}
+			break;
+		case OPT_TOKEN_SECRET:
+			token_str = keep_config_arg();
+			break;
+		case OPT_OS:
+			if (openconnect_set_reported_os(vpninfo, config_arg)) {
+				fprintf(stderr, _("Invalid OS identity \"%s\"\n"),
+					config_arg);
+				exit(1);
+			}
+			break;
 		default:
 			usage();
 		}
@@ -724,11 +782,14 @@ int main(int argc, char **argv)
 #endif
 	}
 
+	if (token_mode != OC_TOKEN_MODE_NONE)
+		init_token(vpninfo, token_mode, token_str);
+
 	if (proxy && openconnect_set_http_proxy(vpninfo, strdup(proxy)))
 		exit(1);
 
 	if (use_syslog) {
-#ifndef ANDROID
+#ifndef __ANDROID__
 		openlog("openconnect", LOG_PID, LOG_DAEMON);
 #endif
 		vpninfo->progress = syslog_progress;
@@ -780,7 +841,7 @@ int main(int argc, char **argv)
 	if (cookieonly == 3) {
 		/* --authenticate */
 		printf("COOKIE='%s'\n", vpninfo->cookie);
-		printf("HOST='%s'\n", vpninfo->hostname);
+		printf("HOST='%s'\n", openconnect_get_hostname(vpninfo));
 		if (vpninfo->peer_cert) {
 			char buf[41] = {0, };
 			openconnect_get_cert_sha1(vpninfo, vpninfo->peer_cert, buf);
@@ -820,8 +881,8 @@ int main(int argc, char **argv)
 	vpn_progress(vpninfo, PRG_INFO,
 		     _("Connected %s as %s%s%s, using %s\n"), vpninfo->ifname,
 		     vpninfo->vpn_addr?:"",
-		     (vpninfo->vpn_addr6 && vpninfo->vpn_addr)?" + ":"",
-		     vpninfo->vpn_addr6?:"",
+		     (vpninfo->vpn_addr6 && vpninfo->vpn_addr) ? " + " : "",
+		     vpninfo->vpn_addr6 ? : "",
 		     (vpninfo->dtls_fd == -1) ?
 		     (vpninfo->deflate ? "SSL + deflate" : "SSL")
 		     : "DTLS");
@@ -881,14 +942,15 @@ static int write_new_config(void *_vpninfo, char *buf, int buflen)
 	}
 
 	/* FIXME: We should actually write to a new tempfile, then rename */
-	if(write(config_fd, buf, buflen) != buflen) {
+	if (write(config_fd, buf, buflen) != buflen) {
 		err = errno;
 		fprintf(stderr, _("Failed to write config to %s: %s\n"),
 			vpninfo->xmlconfig, strerror(err));
-
+		close(config_fd);
 		return -err;
 	}
 
+	close(config_fd);
 	return 0;
 }
 
@@ -908,15 +970,15 @@ void write_progress(void *_vpninfo, int level, const char *fmt, ...)
 	}
 }
 
-#ifdef ANDROID
+#ifdef __ANDROID__
 void syslog_progress(void *_vpninfo, int level, const char *fmt, ...)
 {
-        static int l[4] = {
+	static int l[4] = {
 		ANDROID_LOG_ERROR,	/* PRG_ERR   */
 		ANDROID_LOG_INFO,	/* PRG_INFO  */
 		ANDROID_LOG_DEBUG,	/* PRG_DEBUG */
 		ANDROID_LOG_DEBUG	/* PRG_TRACE */
-        };
+	};
 	va_list args, args2;
 
 	if (verbose >= level) {
@@ -930,7 +992,7 @@ void syslog_progress(void *_vpninfo, int level, const char *fmt, ...)
 		va_end(args2);
 	}
 }
-#else /* !ANDROID */
+#else /* !__ANDROID__ */
 void syslog_progress(void *_vpninfo, int level, const char *fmt, ...)
 {
 	int priority = level ? LOG_INFO : LOG_NOTICE;
@@ -1025,13 +1087,13 @@ static int process_auth_form(void *_vpninfo,
 	char response[1024];
 	char *p;
 
-	if (form->banner)
+	if (form->banner && verbose > PRG_ERR)
 		fprintf(stderr, "%s\n", form->banner);
 
 	if (form->error)
 		fprintf(stderr, "%s\n", form->error);
 
-	if (form->message)
+	if (form->message && verbose > PRG_ERR)
 		fprintf(stderr, "%s\n", form->message);
 
 	/* scan for select options first so they are displayed first */
@@ -1121,7 +1183,7 @@ static int process_auth_form(void *_vpninfo,
 					     _("User input required in non-interactive mode\n"));
 				goto err;
 			} else {
-				opt->value=malloc(80);
+				opt->value = malloc(80);
 				if (!opt->value)
 					goto err;
 
@@ -1149,7 +1211,7 @@ static int process_auth_form(void *_vpninfo,
 				goto err;
 			} else {
 				struct termios t;
-				opt->value=malloc(80);
+				opt->value = malloc(80);
 				if (!opt->value)
 					goto err;
 
@@ -1194,3 +1256,56 @@ static int process_auth_form(void *_vpninfo,
 	}
 	return -EINVAL;
 }
+
+static void init_token(struct openconnect_info *vpninfo,
+		       oc_token_mode_t token_mode, const char *token_str)
+{
+	int ret;
+
+	ret = openconnect_set_token_mode(vpninfo, token_mode, token_str);
+
+	switch (token_mode) {
+	case OC_TOKEN_MODE_STOKEN:
+		switch (ret) {
+		case 0:
+			return;
+		case -EINVAL:
+			fprintf(stderr, _("Soft token string is invalid\n"));
+			exit(1);
+		case -ENOENT:
+			fprintf(stderr, _("Can't open ~/.stokenrc file\n"));
+			exit(1);
+		case -EOPNOTSUPP:
+			fprintf(stderr, _("OpenConnect was not built with libstoken support\n"));
+			exit(1);
+		default:
+			fprintf(stderr, _("General failure in libstoken\n"));
+			exit(1);
+		}
+
+		break;
+
+	case OC_TOKEN_MODE_TOTP:
+		switch (ret) {
+		case 0:
+			return;
+		case -EINVAL:
+			fprintf(stderr, _("Soft token string is invalid\n"));
+			exit(1);
+		case -EOPNOTSUPP:
+			fprintf(stderr, _("OpenConnect was not built with liboath support\n"));
+			exit(1);
+		default:
+			fprintf(stderr, _("General failure in liboath\n"));
+			exit(1);
+		}
+
+		break;
+
+	case OC_TOKEN_MODE_NONE:
+		/* No-op */
+		break;
+
+	/* Option parsing already checked for invalid modes. */
+	}
+}
