@@ -3,6 +3,7 @@
  *
  * Copyright © 2008-2012 Intel Corporation.
  * Copyright © 2008 Nick Andrew <nick@nick-andrew.net>
+ * Copyright © 2013 John Morrissey <jwm@horde.net>
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -28,7 +29,7 @@
 
 #include "openconnect.h"
 
-#if defined (OPENCONNECT_OPENSSL) || defined(DTLS_OPENSSL)
+#if defined(OPENCONNECT_OPENSSL) || defined(DTLS_OPENSSL)
 #include <openssl/ssl.h>
 #include <openssl/err.h>
 /* Ick */
@@ -39,7 +40,7 @@
 #endif
 #endif /* OPENSSL */
 
-#if defined (OPENCONNECT_GNUTLS)
+#if defined(OPENCONNECT_GNUTLS)
 #include <gnutls/gnutls.h>
 #include <gnutls/abstract.h>
 #include <gnutls/x509.h>
@@ -61,6 +62,10 @@
 #include LIBPROXY_HDR
 #endif
 
+#ifdef HAVE_LIBSTOKEN
+#include <stoken.h>
+#endif
+
 #ifdef ENABLE_NLS
 #include <locale.h>
 #include <libintl.h>
@@ -70,6 +75,8 @@
 #endif
 #define N_(s) s
 
+#include <libxml/tree.h>
+
 #define SHA1_SIZE 20
 #define MD5_SIZE 16
 
@@ -123,10 +130,16 @@ struct pin_cache {
 #define CERT_TYPE_PKCS12	2
 #define CERT_TYPE_TPM		3
 
+#define REDIR_TYPE_NONE		0
+#define REDIR_TYPE_NEWHOST	1
+#define REDIR_TYPE_LOCAL	2
+
 struct openconnect_info {
 	char *redirect_url;
+	int redirect_type;
 
 	const char *csd_xmltag;
+	const char *platname;
 	char *csd_token;
 	char *csd_ticket;
 	char *csd_stuburl;
@@ -135,6 +148,7 @@ struct openconnect_info {
 	char *csd_preurl;
 
 	char *csd_scriptname;
+	xmlNode *opaque_srvdata;
 
 #ifdef LIBPROXY_HDR
 	pxProxyFactory *proxy_factory;
@@ -145,6 +159,7 @@ struct openconnect_info {
 
 	const char *localname;
 	char *hostname;
+	char *unique_hostname;
 	int port;
 	char *urlpath;
 	int cert_expire_warning;
@@ -160,11 +175,26 @@ struct openconnect_info {
 	char *password;
 	char *authgroup;
 	int nopasswd;
+	int xmlpost;
 	char *dtls_ciphers;
 	uid_t uid_csd;
 	char *csd_wrapper;
 	int uid_csd_given;
 	int no_http_keepalive;
+	int dump_http_traffic;
+
+	int token_mode;
+	int token_bypassed;
+	int token_tries;
+	time_t token_time;
+#ifdef HAVE_LIBSTOKEN
+	struct stoken_ctx *stoken_ctx;
+	char *stoken_pin;
+#endif
+#ifdef HAVE_LIBOATH
+	char *oath_secret;
+	size_t oath_secret_len;
+#endif
 
 	OPENCONNECT_X509 *peer_cert;
 
@@ -289,8 +319,8 @@ struct openconnect_info {
 	openconnect_progress_vfn progress;
 };
 
-#if (defined (DTLS_OPENSSL) && defined (SSL_OP_CISCO_ANYCONNECT)) || \
-    (defined (DTLS_GNUTLS) && defined (HAVE_GNUTLS_SESSION_SET_PREMASTER))
+#if (defined(DTLS_OPENSSL) && defined(SSL_OP_CISCO_ANYCONNECT)) || \
+    (defined(DTLS_GNUTLS) && defined(HAVE_GNUTLS_SESSION_SET_PREMASTER))
 #define HAVE_DTLS 1
 #endif
 
@@ -304,7 +334,7 @@ struct openconnect_info {
 #define AC_PKT_COMPRESSED	8	/* Compressed data */
 #define AC_PKT_TERM_SERVER	9	/* Server kick */
 
-#define vpn_progress(vpninfo, ...) (vpninfo)->progress ((vpninfo)->cbdata, __VA_ARGS__)
+#define vpn_progress(vpninfo, ...) (vpninfo)->progress((vpninfo)->cbdata, __VA_ARGS__)
 
 /****************************************************************************/
 /* Oh Solaris how we hate thee! */
@@ -320,6 +350,18 @@ int openconnect__asprintf(char **strp, const char *fmt, ...);
 #define getline openconnect__getline
 ssize_t openconnect__getline(char **lineptr, size_t *n, FILE *stream);
 #endif
+#ifndef HAVE_STRCASESTR
+#define strcasestr openconnect__strcasestr
+char *openconnect__strcasestr(const char *haystack, const char *needle);
+#endif
+
+/* I always coded as if it worked like this. Now it does. */
+#define realloc_inplace(p, size) do {			\
+	void *__realloc_old = p;			\
+	p = realloc(p, size);				\
+	if (size && !p)					\
+		free(__realloc_old);			\
+    } while (0)
 
 /****************************************************************************/
 
@@ -327,7 +369,7 @@ ssize_t openconnect__getline(char **lineptr, size_t *n, FILE *stream);
 int setup_tun(struct openconnect_info *vpninfo);
 int tun_mainloop(struct openconnect_info *vpninfo, int *timeout);
 void shutdown_tun(struct openconnect_info *vpninfo);
-int script_config_tun (struct openconnect_info *vpninfo, const char *reason);
+int script_config_tun(struct openconnect_info *vpninfo, const char *reason);
 
 /* dtls.c */
 unsigned char unhex(const char *data);
@@ -350,11 +392,11 @@ int  __attribute__ ((format (printf, 2, 3)))
     openconnect_SSL_printf(struct openconnect_info *vpninfo, const char *fmt, ...);
 int openconnect_print_err_cb(const char *str, size_t len, void *ptr);
 #define openconnect_report_ssl_errors(v) ERR_print_errors_cb(openconnect_print_err_cb, (v))
-#ifdef FAKE_ANDROID_KEYSTORE
+#if defined(FAKE_ANDROID_KEYSTORE) || defined(__ANDROID__)
 #define ANDROID_KEYSTORE
 #endif
 #ifdef ANDROID_KEYSTORE
-char *keystore_strerror(int err);
+const char *keystore_strerror(int err);
 int keystore_fetch(const char *key, unsigned char **result);
 #endif
 
@@ -370,6 +412,11 @@ int openconnect_sha1(unsigned char *result, void *data, int len);
 int openconnect_random(void *bytes, int len);
 int openconnect_local_cert_md5(struct openconnect_info *vpninfo,
 			       char *buf);
+#if defined(OPENCONNECT_OPENSSL)
+#define openconnect_https_connected(_v) ((_v)->https_ssl)
+#elif defined (OPENCONNECT_GNUTLS)
+#define openconnect_https_connected(_v) ((_v)->https_sess)
+#endif
 
 /* mainloop.c */
 int vpn_add_pollfd(struct openconnect_info *vpninfo, int fd, short events);
@@ -386,8 +433,13 @@ int config_lookup_host(struct openconnect_info *vpninfo, const char *host);
 
 /* auth.c */
 int parse_xml_response(struct openconnect_info *vpninfo, char *response,
-		       char *request_body, int req_len, const char **method,
-		       const char **request_body_type);
+		       struct oc_auth_form **form, int *cert_rq);
+int handle_auth_form(struct openconnect_info *vpninfo, struct oc_auth_form *form,
+		     char *request_body, int req_len, const char **method,
+		     const char **request_body_type);
+void free_auth_form(struct oc_auth_form *form);
+int xmlpost_initial_req(struct openconnect_info *vpninfo, char *request_body, int req_len, int cert_fail);
+int prepare_stoken(struct openconnect_info *vpninfo);
 
 /* http.c */
 char *openconnect_create_useragent(const char *base);
@@ -398,10 +450,6 @@ int internal_parse_url(char *url, char **res_proto, char **res_host,
 /* ssl_ui.c */
 int set_openssl_ui(void);
 
-/* securid.c */
-int generate_securid_tokencodes(struct openconnect_info *vpninfo);
-int add_securid_pin(char *token, char *pin);
-
 /* version.c */
 extern const char *openconnect_version_str;
 
